﻿using Microwave.Classes.Boundary.Interfaces;
using Microwave.Classes.Controllers;
using Microwave.Classes.Controllers.Interfaces;
using NSubstitute;

namespace Microwave.Classes.Test.ControllerTest;

[TestFixture]
public class CookControllerTest
{
    private CookController _uut;

    private IMicrowaveController _ui;
    private ITimeProvider _timer;
    private IDisplay _display;
    private IPowerTube _powerTube;

    [SetUp]
    public void Setup()
    {
        _ui = Substitute.For<IMicrowaveController>();
        _timer = Substitute.For<ITimeProvider>();
        _display = Substitute.For<IDisplay>();
        _powerTube = Substitute.For<IPowerTube>();
        _powerTube.GetMaxPower().Returns(700);

        _uut = new CookController(_timer, _display, _powerTube, _ui);
    }

    [Test]
    public void StartCooking_ValidParameters_TimerStarted()
    {
        _uut.StartCooking(50, 60);

        _timer.Received().Start(60);
    }

    [Test]
    public void StartCooking_ValidParameters_PowerTubeStarted()
    {
        _uut.StartCooking(50, 60);

        _powerTube.Received().TurnOn(50);
    }

    [Test]
    public void Cooking_TimerTick_DisplayCalled()
    {
        _uut.StartCooking(50, 60);

        _timer.TimeRemaining.Returns(115);
        _timer.TimerTick += Raise.EventWith(this, EventArgs.Empty);

        _display.Received().ShowTime(1, 55);
    }

    [Test]
    public void NotCooking_TimerTick_DisplayNotCalled()
    {
        // Arrange
        _uut.Stop();

        // Act
        _timer.TimerTick += Raise.EventWith(this, EventArgs.Empty);

        // Assert
        _display.DidNotReceive().ShowTime(default, default);
    }

    [Test]
    public void Cooking_ExtendCookingTime_TimerExtended()
    {
        _uut.StartCooking(50, 60);
        _uut.ExtendCookingTime();

        _timer.Received().AddTime();
    }

    [Test]
    public void NotCooking_ExtendCookingTime_TimerNotExtended()
    {
        _uut.StartCooking(50, 60);
        _uut.Stop();
        _uut.ExtendCookingTime();

        _timer.DidNotReceive().AddTime();
    }

    [Test]
    public void Cooking_TimerExpired_PowerTubeOff()
    {
        _uut.StartCooking(50, 60);

        _timer.Expired += Raise.EventWith(this, EventArgs.Empty);

        _powerTube.Received().TurnOff();
    }

    [Test]
    public void Cooking_TimerExpired_UICalled()
    {
        _uut.StartCooking(50, 60);

        _timer.Expired += Raise.EventWith(this, EventArgs.Empty);

        _ui.Received().CookingIsDone();
    }

    [Test]
    public void NotCooking_TimerExpired_NothingHappens()
    {
        // Arrange
        _uut.Stop();

        // Act
        _timer.Expired += Raise.EventWith(this, EventArgs.Empty);

        // Assert
        _powerTube.Received(1).TurnOff(); // It gets turned off in Stop() Therefore we check that it is called only once
        _ui.DidNotReceive().CookingIsDone();
    }

    [Test]
    public void Cooking_Stop_PowerTubeOff()
    {
        _uut.StartCooking(50, 60);
        _uut.Stop();

        _powerTube.Received().TurnOff();
    }

    [Test]
    public void GetPowerTubeMaxPower_CorrectValue()
    {
        Assert.That(_uut.GetPowerTubeMaxPower() == 700);
    }

    [Test]
    public void OnTimerExpired_IsNotCooking_ReturnImmediately()
    {
        // Arrange
        // Ensure isCooking is false (it's false by default after construction)

        // Act
        _timer.Expired += Raise.EventWith(this, EventArgs.Empty);

        // Assert
        // Verify that the PowerTube was NOT turned off (meaning the method returned early)
        _powerTube.DidNotReceive().TurnOff();
    }

    [Test]
    public void OnTimerTick_IsNotCooking_ReturnImmediately()
    {
        // Arrange

        // Act
        _timer.TimerTick += Raise.EventWith(this, EventArgs.Empty);

        // Assert
        // Verify that the Display was NOT updated (meaning the method returned early)
        _display.DidNotReceive().ShowTime(Arg.Any<int>(), Arg.Any<int>());
    }
}
